@@ -8,7 +8,7 @@ from statistics import mean
 from basics.base import Base
 import basics.base_utils as _
 
-from mlpug.utils import convert_to_dict, get_value_at, has_key, SlidingWindow
+from mlpug.utils import convert_to_dict, get_value_at, set_value_at, has_key, SlidingWindow
 
 from mlpug.mlpug_exceptions import TrainerInvalidException
 
@@ -66,8 +66,7 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
         self._previous_batch_duration = None
 
-        self.training_loss_window = None
-        self.batch_duration_window = None
+        self._metric_windows = {}
 
         self._stop_training = False
 
@@ -79,8 +78,11 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
         self._assess_num_batches_per_epoch()
 
         if type(self.num_batches_per_epoch) == int and self.num_batches_per_epoch > 0:
-            self.training_loss_window = SlidingWindow(length=self.num_batches_per_epoch)
-            self.batch_duration_window = SlidingWindow(length=self.num_batches_per_epoch)
+            self._metric_windows['training.batch.loss'] = \
+                SlidingWindow(length=self.num_batches_per_epoch, name='training.batch.loss')
+
+            self._metric_windows['training_params.batch.duration'] = \
+                SlidingWindow(length=self.num_batches_per_epoch, name='training_params.batch.duration')
 
         if not self._call_callbacks("set_training_manager", self):
             self._callback_calls_successful = False
@@ -123,8 +125,7 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                 "batch_step": self.batch_step,
                 "global_iter": self.global_iter,
                 "logs": self.logs,
-                "training_loss_window": self.training_loss_window.get_state() if self.training_loss_window else None,
-                "batch_duration_window": self.batch_duration_window.get_state() if self.batch_duration_window else None
+                "metric_windows": self._get_metric_windows_state()
             },
 
             "callbacks": {},
@@ -141,6 +142,12 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
                 success &= cb_success
 
+                if cb_name in state["callbacks"] and cb_state != state["callbacks"][cb_name]:
+                    self._log.error(f"There is already a callback in the TrainingManager state "
+                                    f"with the name {cb_name}, this state will be overridden now. "
+                                    f"Please ensure that all your callbacks have a unique name")
+                    success = False
+
                 state["callbacks"][cb_name] = cb_state
             except Exception as e:
                 _.log_exception(self._log, f"Failed to get state of callback {cb_name}, "
@@ -163,6 +170,7 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                    "batch_step": self.batch_step,
                    "global_iter": self.global_iter,
                    "logs": self.logs,
+                   "experiment_data": self.experiment_data
                }, True
 
     def set_state(self, state):
@@ -181,19 +189,39 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
             self._log.error("Invalid state object, unable to set state")
             return False
 
+        self.init_logs = state["manager"]["logs"]
+
         self.epoch = state["manager"]["epoch"]
         self.batch_step = state["manager"]["batch_step"]
         self.global_iter = state["manager"]["global_iter"]
-        self.init_logs = state["manager"]["logs"]
 
-        success = self._set_window_state(state, "training_loss_window")
-        success &= self._set_window_state(state, "batch_duration_window")
+        # start at next iteration
+        self.global_iter += 1
+        self.batch_step += 1
+        if self.batch_step > self.num_batches_per_epoch-1:
+            self.epoch += 1
+            self.batch_step = 0
+
+        self.init_logs["epoch"] = self.epoch
+        self.init_logs["batch_step"] = self.batch_step
+        self.init_logs["global_iter"] = self.global_iter
+
+        success = self._set_metric_windows_states(state)
+
+        # TODO : legacy from <= 0.0.14
+        if "training_loss_window" in state["manager"]:
+            self._log.debug("Loading deprecated metric window state training_loss_window ...")
+            self._set_metric_window_state("training.batch.loss", state["manager"]["training_loss_window"])
+
+        if "batch_duration_window" in state["manager"]:
+            self._log.debug("Loading deprecated metric window state batch_duration_window ...")
+            self._set_metric_window_state("training_params.batch.duration", state["manager"]["batch_duration_window"])
 
         try:
             callbacks_map = {str(callback): callback for callback in self.callbacks}
 
-            callbacks = state["callbacks"] or {}
-            for cb_name, cb_state in callbacks.items():
+            callbacks_state = state["callbacks"] or {}
+            for cb_name, cb_state in callbacks_state.items():
                 if cb_name not in callbacks_map:
                     self._log.warn(f"Callback {cb_name} not given, unable to set callback state, skipping ...")
                     continue
@@ -217,23 +245,46 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
         return success
 
-    def _set_window_state(self, state, window_name):
+    def _get_metric_windows_state(self):
+        state = {}
+        for metric_path, metric_window in self._metric_windows.items():
+            state[metric_path] = metric_window.get_state()
+
+        return state
+
+    def _set_metric_windows_states(self, state):
         success = True
         try:
-            window_state = state["manager"][window_name]
-            if window_state:
-                window_length = window_state['length']
+            metric_windows_state = get_value_at("manager.metric_windows", state)
+            if metric_windows_state is None:
+                return success
 
-                if self.num_batches_per_epoch and (window_length != self.num_batches_per_epoch):
-                    # override when different number of batches per epoch is given (or calculated) during construction
-                    window = SlidingWindow(length=self.num_batches_per_epoch,
-                                           init_window_values=window_state['window'])
-                else:
-                    window = SlidingWindow(state=window_state)
+            for metric_path, window_state in metric_windows_state.items():
+                success &= self._set_metric_window_state(metric_path, window_state)
 
-                setattr(self, window_name, window)
         except Exception as e:
-            _.log_exception(self._log, f"Unable to set training loss window state, skipped ...", e)
+            _.log_exception(self._log, f"Unable to set metric windows state, skipped ...", e)
+            success = False
+
+        return success
+
+    def _set_metric_window_state(self, metric_path, window_state):
+        success = True
+        try:
+            window_length = window_state['length']
+
+            if self.num_batches_per_epoch and (window_length != self.num_batches_per_epoch):
+                # override when different number of batches per epoch is given (or calculated)
+                # during construction
+                window = SlidingWindow(length=self.num_batches_per_epoch,
+                                       init_window_values=window_state['window'],
+                                       name=metric_path)
+            else:
+                window = SlidingWindow(state=window_state)
+
+            self._metric_windows[metric_path] = window
+        except Exception as e:
+            _.log_exception(self._log, f"Unable to set metric window state for {metric_path}, skipped ...", e)
             success = False
 
         return success
@@ -244,14 +295,20 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                             'manager.batch_step',
                             'manager.global_iter',
                             'manager.logs',
-                            'manager.training_loss_window',
-                            'manager.batch_duration_window',
+                            'manager.metric_windows',
                             'callbacks',
                             'trainer']
 
         for attr in state_attributes:
             v = get_value_at(attr, state, warn_on_failure=False)
+
             if v is None:
+                if attr == "manager.metric_windows":
+                    v = get_value_at('manager.training_loss_window', state, warn_on_failure=False)
+                    if v is not None:
+                        self._log.debug("Legacy state detected (training_loss_window) ... ")
+                        continue
+
                 self._log.error(f"Given state does not have a value for {attr}, state is invalid")
                 return False
 
@@ -332,8 +389,8 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                     ctl["batch"]["loss"], ctl["batch"]["auxiliary_results"] = self.trainer.train_on(training_batch,
                                                                                                     training_settings)
 
-                    self._update_window("training_loss_window", ctl["batch"]["loss"])
-                    ctl["window_average"]["loss"] = self._calc_window_average("training_loss_window")
+                    self._update_window("training.batch.loss")
+                    ctl["window_average"]["loss"] = self._calc_window_average("training.batch.loss")
                 except Exception as e:
                     if isinstance(e, TrainerInvalidException):
                         err_msg = f"Trainer {self.trainer} is misconfigured, unable to train on batch, " \
@@ -358,7 +415,10 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                 batch_training_end_time = time.time()
 
                 batch_duration = batch_training_end_time - batch_training_start_time
-                self._update_window("batch_duration_window", batch_duration)
+
+                set_value_at("training_params.batch.duration", current, batch_duration)
+
+                self._update_window("training_params.batch.duration")
 
                 self._previous_batch_duration = batch_duration
 
@@ -371,8 +431,8 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
             if not (epoch_stopped_early or training_stopped_on_error):
                 ctpl = current["training_params"]
 
-                ctpl["window_average"]["duration"] = self._calc_window_average("batch_duration_window")
-                ctpl["epoch"]["duration"] = self._calc_window_sum("batch_duration_window")
+                ctpl["window_average"]["duration"] = self._calc_window_average("training_params.batch.duration")
+                ctpl["epoch"]["duration"] = self._calc_window_sum("training_params.batch.duration")
 
                 logs["cb_calls_success"] &= self._call_callbacks('on_epoch_completed', logs)
 
@@ -406,7 +466,7 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
             "training": {
                 "batch": {},
                 "window_average": {
-                    "loss": self._calc_window_average("training_loss_window")
+                    "loss": self._calc_window_average("training.batch.loss")
                 },
                 "dataset": {},
             },
@@ -416,10 +476,10 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                     "duration": self._previous_batch_duration,
                 },
                 "window_average": {
-                    "duration": self._calc_window_average("batch_duration_window")
+                    "duration": self._calc_window_average("training_params.batch.duration")
                 },
                 "epoch": {
-                    "duration": self._calc_window_sum("batch_duration_window")
+                    "duration": self._calc_window_sum("training_params.batch.duration")
                 }
             }
         }
@@ -444,38 +504,40 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
         return success
 
-    def _update_window(self, window_name, value):
+    def _update_window(self, metric_path):
         try:
-            window_obj = getattr(self, window_name)
-            if not window_obj:
+            window = self._metric_windows[metric_path]
+            if window is None:
                 return
 
-            window_obj.slide(value)
-        except Exception as e:
-            _.log_exception(self._log, f"Exception occurred updating sliding window {window_name}, skipped...", e)
+            value = get_value_at(metric_path, self.logs["current"])
 
-    def _calc_window_average(self, window_name):
+            window.slide(value)
+        except Exception as e:
+            _.log_exception(self._log, f"Exception occurred updating sliding window {metric_path}, skipped...", e)
+
+    def _calc_window_average(self, metric_path):
         try:
-            window_obj = getattr(self, window_name)
-            if not window_obj:
+            window = self._metric_windows[metric_path]
+            if window is None:
                 return None
 
-            window_data = getattr(self, window_name).window
+            window_data = window.window
             return mean(window_data) if len(window_data) > 0 else None
         except Exception as e:
-            _.log_exception(self._log, f"Exception occurred calculating sliding window average for {window_name}.", e)
+            _.log_exception(self._log, f"Exception occurred calculating sliding window average for {metric_path}.", e)
             return None
 
-    def _calc_window_sum(self, window_name):
+    def _calc_window_sum(self, metric_path):
         try:
-            window_obj = getattr(self, window_name)
-            if not window_obj:
+            window = self._metric_windows[metric_path]
+            if window is None:
                 return None
 
-            window_data = getattr(self, window_name).window
+            window_data = window.window
             return sum(window_data) if len(window_data) > 0 else 0
         except Exception as e:
-            _.log_exception(self._log, f"Exception occurred calculating sliding window average for {window_name}.", e)
+            _.log_exception(self._log, f"Exception occurred calculating sliding window average for {metric_path}.", e)
             return None
 
     def _prepare_training_dataset(self):
@@ -659,9 +721,44 @@ class TrainerBase(Base, metaclass=abc.ABCMeta):
 
         return success
 
+    def set_learning_rate(self, lr):
+        """
+        Convenience method to set the learning rate of all optimizers to `lr`
+
+        :param lr: Learning rate to set
+        :return: True on success else False
+        """
+        success = True
+
+        optimizer_names = self.get_optimizers().keys()
+        for opt_name in optimizer_names:
+            try:
+                success &= self.set_learning_rate_for(opt_name, lr)
+            except Exception as e:
+                _.log_exception(self._log, f"Unable to set learning rate for optimizer {opt_name}", e)
+                success = False
+
+        return success
+
+    @abc.abstractmethod
+    def set_learning_rate_for(self, optimizer_name, lr):
+        """
+
+        Set learning rate for specific optimizer `optimizer_name` to `lr`
+
+        :param optimizer_name:
+        :param lr:
+
+        :return: True on success, else False
+        """
+        pass
+
     @abc.abstractmethod
     def train_on(self, batch_data, training_settings=None):
         """
+        TODO : Should this also return only one dict with 'loss' and 'auxiliary_results' keys?
+               (Just like evaluate_loss)
+
         Use batch_data to perform a training iteration
 
         :param batch_data: batch_data object (e.g. dict, list, tuple)
@@ -682,10 +779,12 @@ class TrainerBase(Base, metaclass=abc.ABCMeta):
                                If False the loss will be evaluated in training mode
         :param evaluate_settings: optional evaluate_settings object (usually dict)
 
-        :return: loss, auxiliary_results
+        :return: dict:
+            {
+                "loss": <Tensor>,
+                "auxiliary_results": <can be anything, e.g dict or list with values or data items>
+            }
 
-        loss : Tensor
-        auxiliary_results : can be anything, e.g dict or list with values or data items
         """
 
         self._activate_inference_mode(inference_mode)
@@ -699,10 +798,11 @@ class TrainerBase(Base, metaclass=abc.ABCMeta):
         :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
         :param evaluate_settings: optional evaluate_settings object (usually dict)
 
-        :return: loss, auxiliary_results
-
-        loss : Tensor
-        auxiliary_results : can be anything, e.g dict or list with values or data items
+        :return: dict:
+            {
+                "loss": <Tensor>,
+                "auxiliary_results": <can be anything, e.g dict or list with values or data items>
+            }
         """
 
         pass
@@ -747,7 +847,11 @@ class TrainerBase(Base, metaclass=abc.ABCMeta):
 
 class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
 
-    def __init__(self, optimizers, model_components=None, batch_chunk_size=None):
+    def __init__(self,
+                 optimizers,
+                 model_components=None,
+                 batch_chunk_size=None,
+                 use_mixed_precision=False):
         """
         Simple trainer based on a training_model, that evaluates the loss on batch data
 
@@ -766,6 +870,8 @@ class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
                                  Note 2.
                                  When using chunked batch processing, the default implementation assumes that the
                                  loss, calculated over a chunk, is the average of the sample losses
+
+        :param use_mixed_precision: If True, Float16/Float32 mixed precision is applied.
         """
 
         model_components = convert_to_dict("model", model_components)
@@ -777,13 +883,21 @@ class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
         if self.batch_chunk_size is not None:
             self._log.info(f"Will train on batches by slicing the batch is chunks of {batch_chunk_size} samples.")
 
+        self.use_mixed_precision = use_mixed_precision
+        if self.use_mixed_precision:
+            self._log.info(f"Will train with mixed precision : {self.use_mixed_precision}")
+
         self.training_model = None
 
     def set_training_model(self, model):
         """
         :param model: nn.Module that returns the loss based on the given batch
                       The forward method of the training model must be callable and the following signature:
-                      model(self, batch_data, training_settings) -> (loss, auxiliary_results)
+
+                      model(self, batch_data, training_settings) -> {
+                        "loss": <Tensor>,
+                        "auxiliary_results": <can be anything, e.g dict or list with values or data items>
+                      }
 
         :param model:
         :return:
@@ -802,14 +916,14 @@ class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
         :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
         :param evaluate_settings: optional evaluate_settings object (usually dict)
 
-        :return: loss, auxiliary_results
 
-        loss : Tensor
-        auxiliary_results : can be anything, e.g dict or list with values or data items
+        :return: dict:
+            {
+                "loss": <Tensor>,
+                "auxiliary_results": <can be anything, e.g dict or list with values or data items>
+            }
         """
-        loss, auxiliary_results = self.training_model(batch_data, evaluate_settings)
-
-        return loss, auxiliary_results
+        return self.training_model(batch_data, evaluate_settings)
 
     def _validate_model(self):
         # TODO : this is framework dependent
